#include "nr_d.h"
#include "nrutil_d.h"
#include "globalVar.h"
#include <math.h>
#include "matrixTools.h"
#include <stdio.h>

void initMatrixHfixRollResist(int iNumDataSets, double** pfMatrixH, double dLambda);
void initMatrixHfixAirResist(int iNumDataSets, double** pfMatrixH, double dLambda);

/** this function calculates the coefficients a0 and a2 in case of 2x2 Matrices
 *  arguments: pointers to store the coefficients a0 and a2; 
 *  pToVData,pToAData: pointer to vData and aData of the manipulated DataSet
 *  return value: the coefficients are stored in the matrix pfMatrixResult
 **/

void calcCoefficients(double* pfCoeffA0, double* pfCoeffA2, varDoubleArr* pToVData, varDoubleArr* pToAData, double dMass)
{
  double** pfMatrixATA;
  double** pfMatrixATb;
  double** pfMatrixResult;
  double fSumVelSquare = 0;    //sum(v^2)
  double fSumVelPowFour = 0;   //sum(v^4)
  double fSumAcc = 0;          //sum(a)
  double fSumVelSquareAcc = 0; //sum(v^2*a)

  long i;
  int iDimRow, iDimCol;

  pfMatrixATA = matrix(1,2,1,2);
  pfMatrixATb = matrix(1,2,1,1);

  for (i=0;i<pToVData->liArrLength;i++)
  {
    fSumVelSquare += pToVData->pfArr[i] * pToVData->pfArr[i];
    fSumVelPowFour += pow(pToVData->pfArr[i],4);
    fSumAcc += pToAData->pfArr[i];
    fSumVelSquareAcc += pToVData->pfArr[i] * (pToVData->pfArr[i] * pToAData->pfArr[i]);
  }

  pfMatrixATA[1][1] = pToVData->liArrLength;  //N
  pfMatrixATA[1][2] = fSumVelSquare;      //sum(v^2)
  pfMatrixATA[2][1] = fSumVelSquare;      //sum(v^2)
  pfMatrixATA[2][2] = fSumVelPowFour;     //sum(v^4)

  pfMatrixATb[1][1] = dMass * (-1) * fSumAcc;            //sum(a)
  pfMatrixATb[2][1] = dMass * (-1) * fSumVelSquareAcc;   //sum(v^2*a)
  
  //invert Matrix AT A
  invertMatrix(pfMatrixATA,2);
  //multiply (AT A)^-1 (AT b)
  multMatrix(pfMatrixATA,2,2,pfMatrixATb,2,1,&pfMatrixResult,&iDimRow,&iDimCol);
  
  *pfCoeffA0 = pfMatrixResult[1][1];
  *pfCoeffA2 = pfMatrixResult[2][1];

  free_dmatrix(pfMatrixATA,1,2,1,2);
  free_dmatrix(pfMatrixATb,1,2,1,1);
  free_dmatrix(pfMatrixResult,1,2,1,1);
  
  printf("ao: %lf\n", *pfCoeffA0);
  printf("a2: %lf\n", *pfCoeffA2);
}


/** this function calculates the coefficients of 
 *  arguments: pointer to a varDoubleArr in which the Coefficients are stored
 *  index of start and end DataSet, lambda to couple, 
 *  ResistenceCode: 0 - fixed rolling resistence; 1- fixed air resistence;
 *  default: fixed rolling resistence
 *  return value: the coefficients are stored in the double pCoefficients;
 *  0: no error;  1: not possible to invert matrix
 **/

int calcCoupleCoefficients(double* pCoefficients, int iFromDataSet, int iToDataSet, double dLambda, int iResistenceCode)
{
  double** pfMatrixATA;
  double** pfMatrixATb;
  double** pfMatrixResult;
  double** pfMatrixH;

  long i, j, iCount;
  int iNumDataSets = 0;
  int iDimRow, iDimCol;
  double dMass;
  
  iNumDataSets = iToDataSet - iFromDataSet + 1;
 
  pfMatrixATA = matrix(1,iNumDataSets*2,1,iNumDataSets*2);
  pfMatrixATb = matrix(1,iNumDataSets*2,1,1);
  pfMatrixH = matrix(1,iNumDataSets*2,1,iNumDataSets*2);
  
  //initialize H: select fixed rolling resistence or air resistence 
  //default: fixed rolling resistence
  switch (iResistenceCode)
  {
    case 0:   initMatrixHfixRollResist(iNumDataSets, pfMatrixH, dLambda);
              break;
    case 1:   initMatrixHfixAirResist(iNumDataSets, pfMatrixH, dLambda);
              break;
    default:  initMatrixHfixRollResist(iNumDataSets, pfMatrixH, dLambda);
              break;
  }
  
  //initialize ATA
  for (j=1;j<=iNumDataSets*2;j++)
  {
    for (i=1;i<=iNumDataSets*2;i++)
    {
      pfMatrixATA[i][j] = 0;      
    }
  }
  
  //calculate ATA and ATb matrices 
  iCount = 1;
  for (j=iFromDataSet;j<=iToDataSet;j++)
  {
    double fSumVelSquare = 0;    //sum(v^2)
    double fSumVelPowFour = 0;   //sum(v^4)
    double fSumAcc = 0;          //sum(a)
    double fSumVelSquareAcc = 0; //sum(v^2*a)
    
    varDoubleArr* pToVData = &(arrDataSet[j].vData);
    varDoubleArr* pToAData = &(arrDataSet[j].aData);
    dMass = arrDataSet[j].dDynamicMass;

    for (i=0;i<pToVData->liArrLength;i++)
    {
      fSumVelSquare += pToVData->pfArr[i] * pToVData->pfArr[i];
      fSumVelPowFour += pow(pToVData->pfArr[i],4);
      fSumAcc += pToAData->pfArr[i];
      fSumVelSquareAcc += pToVData->pfArr[i] * (pToVData->pfArr[i] * pToAData->pfArr[i]);
    }

    pfMatrixATA[iCount][iCount] = pToVData->liArrLength;  //N
    pfMatrixATA[iCount][iCount+1] = fSumVelSquare;      //sum(v^2)
    pfMatrixATA[iCount+1][iCount] = fSumVelSquare;      //sum(v^2)
    pfMatrixATA[iCount+1][iCount+1] = fSumVelPowFour;     //sum(v^4)

    pfMatrixATb[iCount][1] = dMass * (-1) * fSumAcc;            //sum(a)
    pfMatrixATb[iCount+1][1] = dMass * (-1) * fSumVelSquareAcc;   //sum(v^2*a)
    
    iCount += 2;
  }
  
  //sum ATA and Lambda*H -> into pfMatrixATA
  for (i=1;i<=iNumDataSets*2;i++)
  {
    for (j=1;j<=iNumDataSets*2;j++)
    {
      pfMatrixATA[i][j]=pfMatrixATA[i][j]+pfMatrixH[i][j];
    }
  }


  //invert Matrix AT A
  if (2 == invertMatrix(pfMatrixATA,iNumDataSets*2))
  { 
    free_dmatrix(pfMatrixATA,1,iNumDataSets*2,1,iNumDataSets*2);
    free_dmatrix(pfMatrixATb,1,iNumDataSets*2,1,1);
    free_dmatrix(pfMatrixH,1,iNumDataSets*2,1,iNumDataSets*2); 
    //not possible to invert matrix;
    return 1;
  }
  //multiply (AT A)^-1 (AT b)
  multMatrix(pfMatrixATA,iNumDataSets*2,iNumDataSets*2,pfMatrixATb,iNumDataSets*2,1,&pfMatrixResult,&iDimRow,&iDimCol);
  
  for (i=1;i<=iDimRow;i++)
  {
		*(pCoefficients+i-1)=pfMatrixResult[i][1];
  }

  free_dmatrix(pfMatrixATA,1,iNumDataSets*2,1,iNumDataSets*2);
  free_dmatrix(pfMatrixATb,1,iNumDataSets*2,1,1);
  free_dmatrix(pfMatrixResult,1,iDimRow,1,iDimCol);
  free_dmatrix(pfMatrixH,1,iNumDataSets*2,1,iNumDataSets*2); 
  return 0;
}


/** this function initializes the matrix H in the case of fixed rolling resistence
 *  arguments: number of data sets, pointer to matrix H, lambda to couple
 *  return value: the matrix H is initialized
 **/

void initMatrixHfixRollResist(int iNumDataSets, double** pfMatrixH, double dLambda)
{
  int i, j;

  //initialize the H matrix - fix rolling resistence   
  for (j=1;j<=iNumDataSets*2;j+=2)
  {
    for (i=1;i<=iNumDataSets*2;i++)
    {
      if(i%2 == 0)
      {
        pfMatrixH[j][i] = 0;
      }
      else
      {
        pfMatrixH[j][i] = -1*dLambda;
      }
      pfMatrixH[j+1][i] = 0;
    }
    pfMatrixH[j][j] = (iNumDataSets-1)*dLambda;
  }
}

/** this function initializes the matrix H in the case of fixed air resistence
 *  arguments: number of data sets, pointer to matrix H, lambda to couple
 *  return value: the matrix H is initialized
 **/

void initMatrixHfixAirResist(int iNumDataSets, double** pfMatrixH, double dLambda)
{
  int i, j;

  //initialize the H matrix - fix air resistence  
  for (j=1;j<=iNumDataSets*2;j+=2)
  {
    for (i=1;i<=iNumDataSets*2;i++)
    {
      if(i%2 == 0)
      {
        pfMatrixH[j+1][i] = -1*dLambda;
      }
      else
      {
        pfMatrixH[j+1][i] = 0;
      }
      pfMatrixH[j][i] = 0;
    }
    pfMatrixH[j+1][j+1] = (iNumDataSets-1)*dLambda;
  }
}
